@@ -21,6 +21,8 @@ the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 #include "config.h"
 
 #include <stdio.h>
+#include <ctype.h>
+#include <limits.h>
 #ifdef IBMPC
 #include <alloc.h>
 #endif
@@ -30,19 +32,19 @@ the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 #endif
 
 #include "netgen.h"
-#include "hash.h"
 #include "objlist.h"
 #include "regexp.h"
 #include "dbug.h"
 #include "print.h"
 #include "netfile.h"
+#include "netcmp.h"
+#include "hash.h"
 
 #ifdef TCL_NETGEN
 extern Tcl_Interp *netgeninterp;
 #endif
 
 struct nlist *CurrentCell = NULL;
-#define USE_TAIL_POINTER
 struct objlist *CurrentTail = NULL;
 
 /* shortcut pointer to last-placed object in list; used in netgen.c */
@@ -122,20 +124,21 @@ struct keyvalue *NewKeyValue(void)
 	return (kv);
 }
 
-struct keylist *NewProperty(void)
+struct property *NewProperty(void)
 {
-	struct keylist *kl;
+	struct property *kl;
 	
-	kl = (struct keylist *)CALLOC(1,sizeof(struct keylist));
+	kl = (struct property *)CALLOC(1,sizeof(struct property));
 	if (kl == NULL) Fprintf(stderr,"NewProperty: Core allocation error\n");
+	kl->pin = (char *)NULL;
 	return (kl);
 }
 
-struct valuelist *NewPropValue(void)
+struct valuelist *NewPropValue(int entries)
 {
 	struct valuelist *vl;
 	
-	vl = (struct valuelist *)CALLOC(1,sizeof(struct valuelist));
+	vl = (struct valuelist *)CALLOC(entries, sizeof(struct valuelist));
 	if (vl == NULL) Fprintf(stderr,"NewPropValue: Core allocation error\n");
 	return (vl);
 }
@@ -171,6 +174,7 @@ char *strsave(char *s)
 	else strcpy(p, s);
 	return (p);
 }
+#endif /* DEBUG_GARBAGE */
 
 /* Case-sensitive matching */
 
@@ -179,17 +183,98 @@ int match(char *st1, char *st2)
 	if (0==strcmp(st1,st2)) return(1);
 	else return(0);
 }
-#endif /* DEBUG_GARBAGE */
 
 /* Case-insensitive matching */
 
+/* For case-insensitivity, the alphabet is lower-cased, but	*/
+/* also all common vector delimiters are cast to the same	*/
+/* character so that different vector notations will match.	*/
+
+static char to_lower[256] = {
+	0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
+	10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
+	20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
+	30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
+	40, 41, 42, 43, 44, 45, 46, 47,
+	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+	':', ';', '<', '=', '>', '?', '@',
+	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
+	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
+	'<', '\\', '>', '^', '_', '`',
+	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
+	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
+	'<', '|', '>', '~', 127,
+	0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
+	10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
+	20, 21, 22, 23, 24, 25, 26, 27, 28, 29,
+	30, 31, 32, 33, 34, 35, 36, 37, 38, 39,
+	40, 41, 42, 43, 44, 45, 46, 47,
+	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
+	':', ';', '<', '=', '>', '?', '@',
+	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
+	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
+	'<', '\\', '>', '^', '_', '`',
+	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
+	'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
+	'<', '|', '>', '~', 127
+};
+
 int matchnocase(char *st1, char *st2)
 {
    char *sp1 = st1;
    char *sp2 = st2;
+   char v1 = FALSE, v2 = FALSE;
+
+   /* In case of a property that does not exist in one netlist, matchnocase()
+    * may be passed a null value, so return 0 to indicate a non-match.
+    * *Both* values null will also be treated as a mismatch (debatable
+    * behavior).
+    */
+   if (!sp1 || !sp2) return 0;
+
+   /* Verilog back-slash escaped names should match an equivalent non-
+    * back-slashed name.  (NOTE: This behavior needs to be added to match().)
+    */
+   if ((*sp1 == '\\') && (*sp2 != '\\')) {
+      v1 = TRUE;
+      sp1++;
+   }
+   if ((*sp2 == '\\') && (*sp1 != '\\')) {
+      v2 = TRUE;
+      sp2++;
+   }
 
    while (*sp1 != '\0' && *sp2 != '\0') {
-      if (tolower(*sp1) != tolower(*sp2)) break;
+      if (to_lower[*sp1] != to_lower[*sp2]) break;
+      sp1++;
+      sp2++;
+   }
+   if (v1 && (*sp1 == ' ')) sp1++;
+   if (v2 && (*sp2 == ' ')) sp2++;
+
+   if ((*sp1 != '\0') || (*sp2 != '\0')) return 0;
+   return 1;
+}
+
+/* Case-sensitive matching with file matching */
+
+int matchfile(char *st1, char *st2, int f1, int f2)
+{
+    if (f1 != f2) return 0;
+    else if (strcmp(st1,st2)) return(0);
+    else return(1);
+}
+
+/* Case-insensitive matching with file matching */
+
+int matchfilenocase(char *st1, char *st2, int f1, int f2)
+{
+   char *sp1 = st1;
+   char *sp2 = st2;
+
+   if (f1 != f2) return 0;
+   while (*sp1 != '\0' && *sp2 != '\0') {
+      if (to_lower[*sp1] != to_lower[*sp2]) break;
       sp1++;
       sp2++;
    }
@@ -224,16 +309,19 @@ void PrintMemoryStats(void)
 
 
 #define CELLHASHSIZE 1000
-static struct hashlist *cell_hashtab[CELLHASHSIZE]; 
+static struct hashdict cell_dict;
 
 void InitCellHashTable(void)
 {
-	InitializeHashTable(cell_hashtab, CELLHASHSIZE);
+    hashfunc = hashcase;
+    matchfunc = NULL;
+    matchintfunc = matchfile;
+    InitializeHashTable(&cell_dict, CELLHASHSIZE);
 }
 
 struct nlist *LookupCell(char *s)
 {
-  return((struct nlist *)HashLookup(s, cell_hashtab, CELLHASHSIZE));
+    return((struct nlist *)HashLookup(s, &cell_dict));
 }
 
 /* Similar hash lookup to the above, but will check if the matching	*/
@@ -243,149 +331,227 @@ struct nlist *LookupCell(char *s)
 struct nlist *LookupCellFile(char *s, int f)
 {
    struct nlist *he;
-   char sf[1024];
-   char *sfp;
 
-   sfp = sf + 1023 - strlen(s);
-   if (sfp < sf) return NULL;
-   strcpy(sfp, s);
-   while (1) {
-      he = HashLookup(sfp, cell_hashtab, CELLHASHSIZE);
-      if (he == NULL)
-	 return NULL;
-      else if ((he->file == -1) || (he->file == f))
-	 return he;
-      sfp--;
-      if (sfp < sf) return NULL;
-      *sfp = '_';
-   }
+   if (f == -1) return LookupCell(s);
+   return HashIntLookup(s, f, &cell_dict);
 }
 
-#if 1
-struct nlist *InstallInCellHashTable(char *name)
+struct nlist *InstallInCellHashTable(char *name, int fnum)
 {
   struct hashlist *ptr;
   struct nlist *p;
 
-  if ((ptr = (struct hashlist *)HashLookup(name, cell_hashtab, CELLHASHSIZE))
-      != NULL) 
-    return((struct nlist *)(ptr->ptr));
+  p = LookupCellFile(name, fnum);
+  if (p != NULL) return(p);
+
+  /* It is not present, so add it to list */
 
-  /* else, it is not present, so add it to list */
   p = (struct nlist *)CALLOC(1, sizeof(struct nlist));
   if (p == NULL) return(NULL);
   if ((p->name = strsave(name)) == NULL) goto fail;
-#ifdef HASH_OBJECTS
-  if ((p->objtab = (struct hashlist **)CALLOC(OBJHASHSIZE, 
-	    sizeof(struct hashlist *))) == NULL) goto fail;
-  if ((p->insttab = (struct hashlist **)CALLOC(OBJHASHSIZE, 
-	    sizeof(struct hashlist *))) == NULL) goto fail;
-#endif
+  p->file = fnum;
+  InitializeHashTable(&(p->objdict), OBJHASHSIZE);
+  InitializeHashTable(&(p->instdict), OBJHASHSIZE);
+  InitializeHashTable(&(p->propdict), OBJHASHSIZE);
+  p->permutes = NULL;
 
-  ptr = HashPtrInstall(name, p, cell_hashtab, CELLHASHSIZE);
+  // Hash size 0 indicates to hash function that no binning is being done
+  p->classhash = (*hashfunc)(name, 0);
+
+  ptr = HashIntPtrInstall(name, fnum, p, &cell_dict);
   if (ptr == NULL) return(NULL);
   return(p);
  fail:
   if (p->name != NULL) FREE(p->name);
-#ifdef HASH_OBJECTS
-  if (p->objtab != NULL) FREE(p->objtab);
-  if (p->insttab != NULL) FREE(p->insttab);
-#endif
-#ifdef CPLUSPLUS
-  FREE((char *)p);  /* added by Glenn;  Mass left it in (I don't care!) */
-#else
+  HashKill(&(p->objdict));
+  HashKill(&(p->instdict));
+  RecurseHashTable(&(p->propdict), freeprop);
+  HashKill(&(p->propdict));
   FREE(p);
-#endif
   return(NULL);
 }
 
-#else
-struct nlist *InstallInCellHashTable(char *name)
-{
-  struct hashlist *ptr;
-  struct nlist *p;
-
-  ptr = HashInstall(name, cell_hashtab, CELLHASHSIZE);
-  if (ptr == NULL) return(NULL);
-  if (ptr->ptr == NULL) {
-    p = (struct nlist *)CALLOC(1, sizeof(struct nlist));
-    if (p == NULL) goto fail;
-    ptr->ptr = (struct hashlist *)p;
-    if ((p->name = strsave(name)) == NULL) goto fail;
-#ifdef HASH_OBJECTS
-    p->objtab = (struct hashlist **)CALLOC(OBJHASHSIZE, 
-	    sizeof(struct hashlist *));
-    p->insttab = (struct hashlist **)CALLOC(OBJHASHSIZE, 
-	    sizeof(struct hashlist *));
-    if (p->objtab == NULL || p->insttab == NULL) goto fail;
-#endif
-  }
-  else p = (struct nlist *)(ptr->ptr);
-  return(p);
-
-  fail:
-    HashDelete(name, cell_hashtab, CELLHASHSIZE);
-    return(NULL);
-}
-#endif
-
 /* Rename the cell named "name" to "newname".  Find the hash entry for
  * cell "name", remove it from the hash table, generate a new hash entry
  * for "newname", and set its pointer to the original cell.
  */
 
-void CellRehash(char *name, char *newname)
+void CellRehash(char *name, char *newname, int file)
 {
   struct nlist *tp;
   struct hashlist *ptr;
 
-  tp = LookupCell(name);
+  if (file == -1)
+     tp = LookupCell(name);
+  else
+     tp = LookupCellFile(name, file);
+
   FREE(tp->name);
   tp->name = strsave(newname);
 
-  ptr = HashPtrInstall(newname, (void *)tp, cell_hashtab, CELLHASHSIZE);
+  ptr = HashIntPtrInstall(newname, file, (void *)tp, &cell_dict);
   if (ptr != NULL)
-     HashDelete(name, cell_hashtab, CELLHASHSIZE);
+     HashIntDelete(name, file, &cell_dict);
+
+  // Change the classhash to reflect the new name
+  tp->classhash = (*hashfunc)(newname, 0);
 }
 
 struct nlist *OldCell;
 
-int deleteclass(struct hashlist *p)
+int removeshorted(struct hashlist *p, int file)
 {
    struct nlist *ptr;
-   struct objlist *ob, *lob, *tob;
+   struct objlist *ob, *lob, *nob, *tob;
+   unsigned char shorted;
 
    ptr = (struct nlist *)(p->ptr);
 
+   if ((file != -1) && (ptr->file != file)) return 0;
+
    lob = NULL;
-   for (ob = ptr->cell; ob != NULL; ob = ob->next) {
-      if ((ob->type == FIRSTPIN) && (ob->model != NULL)) {
-	 if (!strcmp(ob->model, OldCell->name)) {
+   for (ob = ptr->cell; ob != NULL;) {
+      nob = ob->next;
+      if ((ob->type == FIRSTPIN) && (ob->model.class != NULL)) {
+	 if ((*matchfunc)(ob->model.class, OldCell->name)) {
+            shorted = (unsigned char)1;
+	    for (tob = nob; tob->type > FIRSTPIN; tob = tob->next) {
+	       if (tob->node != ob->node) {
+	           shorted = (unsigned char)0;
+	           break;
+	       }
+            }
+	    if (shorted == (unsigned char)0) {
+               lob = ob;
+               ob = nob;
+	       continue;
+	    }
+	    HashDelete(ob->instance.name, &(ptr->instdict));
 	    while (1) {
-	       tob = ob->next;
-	       FreeObject(ob);
-	       ob = tob;
-	       if ((ob == NULL) || (ob->type == FIRSTPIN))
-		  break;
+	       FreeObjectAndHash(ob, ptr);
+	       ob = nob;
+	       if (ob == NULL) break;
+	       nob = ob->next;
+	       if (ob->type != PROPERTY && ob->type <= FIRSTPIN) break;
 	    }
 	    if (lob == NULL)
 	       ptr->cell = ob;
 	    else
 	       lob->next = ob;
 	 }
+	 else {
+	    lob = ob;
+	    ob = nob;
+	 }
+      }
+      else {
+         lob = ob;
+         ob = nob;
       }
-      lob = ob;
-      if (ob == NULL) break;
    }
 }
 
+/* Remove shorted instances of class "class" from the database */
+
+void RemoveShorted(char *class, int file)
+{
+   if (file == -1)
+      OldCell = LookupCell(class);
+   else
+      OldCell = LookupCellFile(class, file);
+
+   if (OldCell == NULL) return;
+   RecurseCellFileHashTable(removeshorted, file);
+}
+
+/* Remove instances of a deleted class from the database.	*/
+/* NOTE:  This treats deleted classes as not existing, so it	*/
+/* needs to take care of disconnected ports in the same manner	*/
+/* as flattenInstancesOf() in disconnecting the port of a	*/
+/* parent cell if it connected to nothing other than the	*/
+/* deleted instance.						*/
+
+int deleteclass(struct hashlist *p, int file)
+{
+   struct nlist *ptr;
+   struct objlist *ob, *lob, *nob, *portnode;
+   unsigned char *checknodes;
+   int i;
+
+   ptr = (struct nlist *)(p->ptr);
+
+   if ((file != -1) && (ptr->file != file)) return 0;
+
+   /* Note:  This could be made faster by enumerating all times each
+    * node is used during the full pass, then subtracting each time
+    * the node is deleted in a child, then disconnecting all nodes
+    * that ended up with a zero count.
+    */
+   checknodes = (unsigned char *)CALLOC(ptr->nodename_cache_maxnodenum + 1,
+		sizeof(unsigned char));
+
+   lob = NULL;
+   for (ob = ptr->cell; ob != NULL;) {
+      nob = ob->next;
+      if ((ob->type == FIRSTPIN) && (ob->model.class != NULL)) {
+	 if ((*matchfunc)(ob->model.class, OldCell->name)) {
+	    HashDelete(ob->instance.name, &(ptr->instdict));
+	    while (1) {
+	       if (ob->type >= FIRSTPIN)
+		  checknodes[ob->node] = (unsigned char)1;
+	       FreeObjectAndHash(ob, ptr);
+	       ob = nob;
+	       if (ob == NULL) break;
+	       nob = ob->next;
+	       if (ob->type != PROPERTY && ob->type <= FIRSTPIN) break;
+	    }
+	    if (lob == NULL)
+	       ptr->cell = ob;
+	    else
+	       lob->next = ob;
+	 }
+	 else {
+	    lob = ob;
+	    ob = nob;
+	 }
+      }
+      else {
+         lob = ob;
+         ob = nob;
+      }
+   }
+
+   for (i = 0; i <= ptr->nodename_cache_maxnodenum; i++) {
+      if (checknodes[i] != 0) {
+         portnode = NULL;
+         for (ob = ptr->cell; ob != NULL; ob = ob->next) {
+	    if ((ob->type != PORT) && (portnode == NULL))
+	       break;
+	    else if ((ob->type == PORT) && (ob->node == i))
+	       portnode = ob;
+	    else if ((ob->type >= FIRSTPIN) && (ob->node == i))
+	       break;
+         }
+         if ((ob == NULL) && (portnode != NULL)) {
+	    /* Port became disconnected when child was deleted */
+	    portnode->node = -1;
+         }
+      }
+   }
+   FREE(checknodes);
+}
+
 /* Remove all instances of class "class" from the database */
 
-void ClassDelete(char *class)
+void ClassDelete(char *class, int file)
 {
-   OldCell = LookupCell(class);
+   if (file == -1)
+      OldCell = LookupCell(class);
+   else
+      OldCell = LookupCellFile(class, file);
+
    if (OldCell == NULL) return;
-   RecurseCellHashTable(deleteclass);
+   RecurseCellFileHashTable(deleteclass, file);
 }
 
 /* Find all instances of the cell named "name" in the database, and 	*/
@@ -393,249 +559,207 @@ void ClassDelete(char *class)
 
 char *NewName;
 
-int renameinstances(struct hashlist *p)
+int renameinstances(struct hashlist *p, int file)
 {
    struct nlist *ptr;
    struct objlist *ob, *ob2, *obp;
 
    ptr = (struct nlist *)(p->ptr);
 
+   if ((file != -1) && (ptr->file != file)) return 0;
+
    for (ob = ptr->cell; ob != NULL; ob = ob->next) {
-      if ((ob->type >= FIRSTPIN) && (ob->model != NULL)) {
-	 if (!strcmp(ob->model, OldCell->name)) {
-	    FreeString(ob->model);
-	    ob->model = strsave(NewName);
+      if ((ob->type >= FIRSTPIN) && (ob->model.class != NULL)) {
+	 if ((*matchfunc)(ob->model.class, OldCell->name)) {
+	    FreeString(ob->model.class);
+	    ob->model.class = strsave(NewName);
 	 }
       }
    }
 }
 
-void InstanceRename(char *from, char *to)
+void InstanceRename(char *from, char *to, int file)
 {
-   OldCell = LookupCell(from);
+   if (file == -1)
+      OldCell = LookupCell(from);
+   else
+      OldCell = LookupCellFile(from, file);
+
    if (OldCell == NULL) return;
    NewName = to;
-   RecurseCellHashTable(renameinstances);
+
+   RecurseCellFileHashTable(renameinstances, file);
 }
 
-void CellDelete(char *name)
+/* Delete contents of a hashed property structure */
+
+int freeprop(struct hashlist *p)
 {
-  /* delete all the contents of cell 'name', and remove 'name' from
-     the cell hash table.  NOTE:  this procedure does not care or check
-     if 'name' has been instanced anywhere.  It is assumed that if this
-     is the case, the user will (quickly) define a new cell of that name.
-  */
-  struct objlist *ob, *obnext;
-  struct keylist *pr, *prnext;
-  struct nlist *tp;
+   struct property *prop;
 
-  tp = LookupCell(name);
-  if (tp == NULL) {
-    Printf ("No cell '%s' found.\n", name);
-    return;
-  }
-
-  HashDelete(name, cell_hashtab, CELLHASHSIZE);
-  /* now make sure that we free all the fields of the nlist struct */
-  if (tp->name != NULL) FREE(tp->name);
-#ifdef HASH_OBJECTS
-  if (tp->objtab != NULL) FREE(tp->objtab);
-  if (tp->insttab != NULL) FREE(tp->insttab);
-#endif
-  FreeNodeNames(tp);
-  ob = tp->cell;
-  while (ob != NULL) {
-    obnext = ob->next;
-    FreeObject (ob);
-    ob = obnext;
-  }
-  pr = tp->proplist;
-  while (pr != NULL) {
-    prnext = pr->next;
-    FreeString(pr->key);
-    FREE(pr);
-    pr = prnext;
-  }
+   prop = (struct property *)(p->ptr);
+   if (prop->type == PROP_STRING)
+      if (prop->pdefault.string != NULL)
+	 FREE(prop->pdefault.string);
+   else if (prop->type == PROP_EXPRESSION) {
+      struct tokstack *stackptr, *nptr;
+      stackptr = prop->pdefault.stack;
+      while (stackptr != NULL) {
+	 nptr = stackptr->next;
+	 if (stackptr->toktype == TOK_STRING)
+	    FREE(stackptr->data.string);
+	 FREE(stackptr);
+	 stackptr = nptr;
+      }
+   }
+   FREE(prop->key);
+   FREE(prop);
+   return 1;
 }
 
-void CellRename(char *from, char *to)
+void CellDelete(char *name, int fnum)
 {
-  /* rename cell 'from' to cell 'to', deleting cell 'to'
-     if it already exists.
-  */
-  struct hashlist *ptr;
-  struct nlist *from_cell, *to_cell;
+   /* delete all the contents of cell 'name', and remove 'name' from
+      the cell hash table.  NOTE:  this procedure does not care or check
+      if 'name' has been instanced anywhere.  It is assumed that if this
+      is the case, the user will (quickly) define a new cell of that name.
+   */
+   struct objlist *ob, *obnext;
+   struct nlist *tp;
 
-  from_cell = LookupCell(from);
-  if (from_cell == NULL) {
-    Printf("No cell '%s' found.\n", from);
-    return;
-  }
-
-  to_cell = LookupCell(to);
-  if (to_cell != NULL) {
-    Printf("Cell '%s' already exists.  Deleting current definition of '%s'\n",
-	   to, to);
-    CellDelete(to);
-  }
-
-  ptr = HashInstall(to, cell_hashtab, CELLHASHSIZE);
-  if (ptr == NULL) {
-    Printf("memory allocation error\n");
-    return;
-  }
-  ptr->ptr = (struct hashlist *)from_cell;
-  FREE(from_cell->name);
-  from_cell->name = strsave(to);
-
-  HashDelete(from, cell_hashtab, CELLHASHSIZE);
-}
-
-void CellCopy(char *from, char *to)
-{
-  /* create a copy of cell 'from', in cell 'to' */
-
-  struct nlist *from_cell, *to_cell, *oldcurcell;
-  struct objlist *ob;
-
-  from_cell = LookupCell(from);
-  if (from_cell == NULL) {
-    Printf("No cell '%s' found.\n", from);
-    return;
-  }
-
-  to_cell = LookupCell(to);
-  if (to_cell != NULL) {
-    Printf("Cell '%s' already exists.  Deleting current definition of '%s'\n",
-	   to, to);
-    CellDelete(to);
-  }
-
-  oldcurcell = CurrentCell;
-  CellDef(to);
-  for (ob = from_cell->cell; ob != NULL; ob = ob->next) {
-    struct objlist *newob;
-
-    if ((newob = GetObject()) == NULL) {
-      perror("Failed GetObject in CellCopy");
+   tp = LookupCellFile(name, fnum);
+   if (tp == NULL) {
+      Printf ("No cell '%s' found.\n", name);
       return;
-    }
-    newob->name = (ob->name) ? strsave(ob->name) : NULL;
-    newob->type = ob->type;
-    if (newob->type == PROPERTY)
-       CopyProperties(ob, newob);
-    else {
-       newob->model = (ob->model) ? strsave(ob->model) : NULL;
-       newob->instance = (ob->instance) ? strsave(ob->instance) : NULL;
-    }
-    newob->node = ob->node;
-    newob->next = NULL;
-    AddToCurrentCell(newob);
-  }
-  EndCell();
-  CurrentCell = oldcurcell;
-}
+   }
 
-  
+   HashIntDelete(name, fnum, &cell_dict);
+   /* now make sure that we free all the fields of the nlist struct */
+   if (tp->name != NULL) FREE(tp->name);
+   HashKill(&(tp->objdict));
+   HashKill(&(tp->instdict));
+   RecurseHashTable(&(tp->propdict), freeprop);
+   HashKill(&(tp->propdict));
+   FreeNodeNames(tp);
+   ob = tp->cell;
+   while (ob != NULL) {
+      obnext = ob->next;
+      FreeObject (ob);
+      ob = obnext;
+   }
+}
 
 static int PrintCellHashTableElement(struct hashlist *p)
 {
-  struct nlist *ptr;
+   struct nlist *ptr;
 
-  ptr = (struct nlist *)(p->ptr);
-  if ((TopFile >= 0) && (ptr->file != TopFile)) return 1;
+   ptr = (struct nlist *)(p->ptr);
+   if ((TopFile >= 0) && (ptr->file != TopFile)) return 1;
 
-  if (ptr->class != CLASS_SUBCKT) {
-    /* only print primitive cells if Debug is enabled */
-    if (Debug == 1)  Printf("Cell: %s (instanced %d times); Primitive\n",
+   if ((ptr->class != CLASS_SUBCKT) && (ptr->class != CLASS_MODULE)) {
+      /* only print primitive cells if Debug is enabled */
+      if (Debug == 1)  Printf("Cell: %s (instanced %d times); Primitive\n",
 		       ptr->name, ptr->number);
-  }
-  else if (Debug == 2) {	/* list only */
+      else if (Debug == 3) {	/* list */
 #ifdef TCL_NETGEN
-     Tcl_AppendElement(netgeninterp, ptr->name);
+          Tcl_AppendElement(netgeninterp, ptr->name);
 #else
-     Printf("%s ", ptr->name);
+          Printf("%s ", ptr->name);
 #endif
-  }
-  else
-    Printf("Cell: %s (instanced %d times)\n",ptr->name,ptr->number);
-  return(1);
+      }
+   }
+   else if ((Debug == 2) || (Debug == 3)) {	/* list only */
+#ifdef TCL_NETGEN
+      Tcl_AppendElement(netgeninterp, ptr->name);
+#else
+      Printf("%s ", ptr->name);
+#endif
+   }
+   else
+      Printf("Cell: %s (instanced %d times)\n", ptr->name, ptr->number);
+   return(1);
 }
 
 /* Print the contents of the cell hash table.	*/
 /* if full == 1, print primitive elements also	*/
 /* if full == 2, just output the cell names.	*/
 
-void PrintCellHashTable(int full, char *filename)
+void PrintCellHashTable(int full, int filenum)
 {
-  int total, bins;
-  int OldDebug;
+   int total, bins;
+   int OldDebug;
 
-  if (filename != NULL) {
-     struct nlist *top;
-     top = LookupCell(filename);
-     TopFile = (top) ? top->file : (File + 1);
-  }
-  else TopFile = -1;
+   if ((filenum == -1) && (Circuit1 != NULL) && (Circuit2 != NULL)) {
+      PrintCellHashTable(full, Circuit1->file);
+      PrintCellHashTable(full, Circuit2->file);
+      return;
+   }
 
-  bins  = RecurseHashTable(cell_hashtab, CELLHASHSIZE, CountHashTableBinsUsed);
-  total = RecurseHashTable(cell_hashtab, CELLHASHSIZE, CountHashTableEntries);
-  if (full != 2)
-     Printf("Hash table: %d of %d bins used; %d cells total (%.2f per bin)\n",
+   TopFile = filenum;
+
+   bins  = RecurseHashTable(&cell_dict, CountHashTableBinsUsed);
+   total = RecurseHashTable(&cell_dict, CountHashTableEntries);
+   if (full < 2)
+       Printf("Hash table: %d of %d bins used; %d cells total (%.2f per bin)\n",
 		bins, CELLHASHSIZE, total, (bins == 0) ? 0 :
 		(float)((float)total / (float)bins));
 	
-  OldDebug = Debug;
-  Debug = full;
-  RecurseHashTable(cell_hashtab, CELLHASHSIZE, PrintCellHashTableElement);
-  Debug = OldDebug;
+   OldDebug = Debug;
+   Debug = full;
+   RecurseHashTable(&cell_dict, PrintCellHashTableElement);
+   Debug = OldDebug;
 #ifndef TCL_NETGEN
-  if (full == 2) Printf("\n");
+   if (full >= 2) Printf("\n");
 #endif
 }
 
 struct nlist *FirstCell(void)
 {
-  return((struct nlist *)HashFirst(cell_hashtab, CELLHASHSIZE));
+   return((struct nlist *)HashFirst(&cell_dict));
 }
 
 struct nlist *NextCell(void)
 {
-  return((struct nlist *)HashNext(cell_hashtab, CELLHASHSIZE));
+   return((struct nlist *)HashNext(&cell_dict));
 }
 
 static int ClearDumpedElement(struct hashlist *np)
 {
-	struct nlist *p;
+   struct nlist *p;
 
-	p = (struct nlist *)(np->ptr);
-	p->dumped = 0;
-	return(1);
+   p = (struct nlist *)(np->ptr);
+   p->dumped = 0;
+   return(1);
 }
 
 void ClearDumpedList(void)
 {
-	RecurseHashTable(cell_hashtab, CELLHASHSIZE, ClearDumpedElement);
+   RecurseHashTable(&cell_dict, ClearDumpedElement);
 }
 
-#ifdef CPLUSPLUS
-int (*MagicFunctionPointer)(struct hashlist *np);
-
-int MagicRecurseCellHashTable(void)
-{
-  return RecurseHashTable(cell_hashtab, CELLHASHSIZE, MagicFunctionPointer);
-}
-#else
 int RecurseCellHashTable(int (*foo)(struct hashlist *np))
 {
-  return RecurseHashTable(cell_hashtab, CELLHASHSIZE, foo);
+   return RecurseHashTable(&cell_dict, foo);
 }
-#endif
+
+int RecurseCellFileHashTable(int (*foo)(struct hashlist *, int), int value)
+{
+   return RecurseHashTableValue(&cell_dict, foo, value);
+}
+
+/* Yet another version, passing one parameter that is a pointer */
+
+struct nlist *RecurseCellHashTable2(struct nlist *(*foo)(struct hashlist *,
+	void *), void *pointer)
+{
+   return RecurseHashTablePointer(&cell_dict, foo, pointer);
+}
+
 /************************** WILD-CARD STUFF *******************************/
 		
 char *FixTemplate(char *t)
 {
-	char buffer[200];
+	char buffer[MAX_STR_LEN];
 	char *rstr;
 	int i,j;
 	int InsideBrace;
@@ -840,8 +964,8 @@ struct objlist *ListCat(struct objlist *ls1, struct objlist *ls2)
 #else
 		tmp->name = ls1->name;
 		tmp->type = ls1->type;
-		tmp->model = ls1->model;
-		tmp->instance = ls1->instance;
+		tmp->model.class = ls1->model.class;
+		tmp->instance.name = ls1->instance.name;
 		tmp->node = ls1->node;
 		tmp->next = NULL;
 #endif
@@ -888,7 +1012,7 @@ int ListLength(char *list_template)
 
 
 
-struct objlist *CopyObjList(struct objlist *oldlist)
+struct objlist *CopyObjList(struct objlist *oldlist, unsigned char doforall)
 /* copies list pointed to by oldlist, creating
    a list whose head pointer is returned */
 {
@@ -910,11 +1034,13 @@ struct objlist *CopyObjList(struct objlist *oldlist)
     if (newob->type == PROPERTY)
        CopyProperties(newob, tmp);
     else {
-       if (tmp->model == NULL || IsPort(tmp->type))
-          newob->model = NULL;
+       if (tmp->model.class == NULL || IsPort(tmp))
+          newob->model.class = NULL;
        else
-          newob->model = strsave(tmp->model);
-       newob->instance = (tmp->instance) ? strsave(tmp->instance) : NULL;
+          newob->model.class = strsave(tmp->model.class);
+       newob->flags = tmp->flags;
+       newob->instance.name = (tmp->instance.name) ?
+		strsave(tmp->instance.name) : NULL;
     }
     newob->node = tmp->node;
     newob->next = NULL;
@@ -924,6 +1050,13 @@ struct objlist *CopyObjList(struct objlist *oldlist)
       tail->next = newob;
     tail = newob;
     tmp = tmp->next;
+
+    // If "doforall" is 0, then only copy one object;  otherwise,
+    // copy to the end of the list.
+    if (!doforall) {
+      if ((tmp == NULL) || ((tmp->type <= FIRSTPIN) && (tmp->type != PROPERTY)))
+	 break;
+    }
   }
   return (head);
 }
@@ -938,51 +1071,13 @@ struct objlist *CopyObjList(struct objlist *oldlist)
 
 struct objlist *LookupObject(char *name, struct nlist *WhichCell)
 {
-#ifdef HASH_OBJECTS
-  return((struct objlist *)HashLookup(name, WhichCell->objtab, OBJHASHSIZE));
-#else		
-	struct objlist *tp;
-	
-	if (QuickSearch && CurrentCell == WhichCell) {
-	  tp = LastPlaced;
-	  while (tp != NULL) {
-		if (match(name, tp->name)) return (tp);
-		tp = tp->next;
-	      }
-	}
-	/* if not QuickSearch, or QuickSearch not successful */
-	tp = WhichCell->cell;
-	while (tp != NULL) {
-		if (match(name, tp->name)) return (tp);
-		tp = tp->next;
-	}
-	return (NULL);
-#endif
+   return((struct objlist *)HashLookup(name, &(WhichCell->objdict)));
 }
 
 struct objlist *LookupInstance(char *name, struct nlist *WhichCell)
 /* searches for exact match of instance 'name' in cell 'WhichCell' */
 {
-#ifdef HASH_OBJECTS
-  return((struct objlist *)HashLookup(name, WhichCell->insttab, OBJHASHSIZE));
-#else		
-	struct objlist *tp;
-	
-	if (QuickSearch && CurrentCell == WhichCell) {
-	  tp = LastPlaced;
-	  while (tp != NULL) {
-		if (match(name, tp->instance)) return (tp);
-		tp = tp->next;
-	      }
-	}
-	/* if not QuickSearch, or QuickSearch not successful */
-	tp = WhichCell->cell;
-	while (tp != NULL) {
-		if (match(name, tp->instance)) return (tp);
-		tp = tp->next;
-	}
-	return (NULL);
-#endif
+   return((struct objlist *)HashLookup(name, &(WhichCell->instdict)));
 }
 
 
@@ -995,8 +1090,17 @@ void UpdateNodeNumbers(struct objlist *lst, int from, int to)
 	}
 }
 
-
 void AddToCurrentCell(struct objlist *ob)
+{
+   AddToCurrentCellNoHash(ob);
+
+   /* add to object hash table for this cell */
+   if (CurrentCell != NULL) {
+      HashPtrInstall(ob->name, ob, &(CurrentCell->objdict));
+   }
+}
+
+void AddToCurrentCellNoHash(struct objlist *ob)
 {
   if (CurrentCell == NULL) {
     Fprintf(stderr,"No current cell for ");
@@ -1010,88 +1114,83 @@ void AddToCurrentCell(struct objlist *ob)
     return;
   }
 
-#ifdef USE_TAIL_POINTER
   if (CurrentCell->cell == NULL) CurrentCell->cell = ob;
   else CurrentTail->next = ob;
   CurrentTail = ob;
   ob->next = NULL;
-#else
-  if (CurrentCell->cell == NULL) CurrentCell->cell = ob;
-  else {
-    struct objlist *ob2;
-      
-    ob2 = CurrentCell->cell;
-    while (ob2->next != NULL) ob2 = ob2->next;
-    ob2->next = ob;
-  }
-  ob->next = NULL;
-#endif
-
-#ifdef HASH_OBJECTS
-  /* add to object hash table for this cell */
-  HashPtrInstall(ob->name, ob, CurrentCell->objtab, OBJHASHSIZE);
-#endif
 }
 
 void AddInstanceToCurrentCell(struct objlist *ob)
 {
-#ifdef HASH_OBJECTS
   /* add to instance hash table for this cell */
-  HashPtrInstall(ob->instance, ob, CurrentCell->insttab, OBJHASHSIZE);
-#endif
+  HashPtrInstall(ob->instance.name, ob, &(CurrentCell->instdict));
 }
 
 void FreeObject(struct objlist *ob)
 {
-  /* for now, just free it; later, we will have to remove from hash table */
+  /* This just frees the object record.  Beware of pointer left	*/
+  /* in the objlist hash table.  Hash table records should be	*/
+  /* removed first.						*/
+
   if (ob->name != NULL) FreeString(ob->name);
 
   if (ob->type == PROPERTY) {
-     /* Property records are different in that they use "instance" and	*/
-     /* "model" to save pointers to linked lists.			*/
-     if (ob->model != NULL) {
-	struct keyvalue *kv, *kvnext;
-	kv = (struct keyvalue *)ob->model;
-	while (kv != NULL) {
-	   kvnext = kv->next;
+     if (ob->instance.props != NULL) {
+	struct valuelist *kv;
+        int i;
+	for (i = 0; ; i++) {
+	   kv = &(ob->instance.props[i]);
+	   if (kv->type == PROP_ENDLIST) break;
 	   FreeString(kv->key);
-	   FreeString(kv->value);
-	   FREE(kv);
-	   kv = kvnext;
-	}
-     }
-     if (ob->instance != NULL) {
-	struct valuelist *vl, *vlnext;
-	vl = (struct valuelist *)ob->instance;
-	while (vl != NULL) {
-	   vlnext = vl->next;
-	   FREE(vl);
-	   vl = vlnext;
+	   if (kv->type == PROP_STRING && kv->value.string != NULL)
+	      FreeString(kv->value.string);
+	   else if (kv->type == PROP_EXPRESSION) {
+	      struct tokstack *stackptr, *nptr;
+	      stackptr = kv->value.stack;
+	      while (stackptr != NULL) {
+		 nptr = stackptr->next;
+		 if (stackptr->toktype == TOK_STRING)
+		    FREE(stackptr->data.string);
+		 FREE(stackptr);
+		 stackptr = nptr;
+	      }
+	   }
 	}
+	FREE(ob->instance.props);
      }
   }
   else {
      /* All other records */
-     if (ob->model != NULL) FreeString(ob->model);
-     if (ob->instance != NULL) FreeString(ob->instance);
+     if (ob->instance.name != NULL) FreeString(ob->instance.name);
   }
+  if (ob->model.class != NULL) FreeString(ob->model.class);
   FREE(ob);
 }
 
 
+void FreeObjectAndHash(struct objlist *ob, struct nlist *ptr)
+{
+   HashDelete(ob->name, &(ptr->objdict));
+   FreeObject(ob);
+}
+
 /***************  GENERAL UTILITIES ****************************/
 
-int NumberOfPorts(char *cellname)
+int NumberOfPorts(char *cellname, int file)
 {
   struct nlist *tp;
   struct objlist *ob;
   int ports;
 
-  tp = LookupCell(cellname);
+  if (file == -1)
+    tp = LookupCell(cellname);
+  else
+    tp = LookupCellFile(cellname, file);
+
   if (tp == NULL) return(0);
   ports = 0;
   for (ob = tp->cell; ob != NULL; ob = ob->next)
-    if (IsPort(ob->type)) ports++;
+    if (IsPort(ob)) ports++;
   return(ports);
 }
   
@@ -1106,10 +1205,10 @@ void FreePorts(char *cellname)
   if (ob == NULL) return;
   tp->cell = NULL;
 
-  while (ob && IsPort(ob->type)) {
+  while (ob && IsPort(ob)) {
     obnext = ob->next;
-    if (ob->model != NULL) FreeString(ob->model);
-    if (ob->instance != NULL) FreeString(ob->instance);
+    if (ob->name != NULL) FreeString(ob->name);
+    if (ob->instance.name != NULL) FreeString(ob->instance.name);
     FREE(ob);
     ob = obnext;
   }
@@ -1118,9 +1217,9 @@ void FreePorts(char *cellname)
   oblast = ob;
   while (ob) {
     obnext = ob->next;
-    if (IsPort(ob->type)) {
-       if (ob->model != NULL) FreeString(ob->model);
-       if (ob->instance != NULL) FreeString(ob->instance);
+    if (IsPort(ob)) {
+       if (ob->name != NULL) FreeString(ob->name);
+       if (ob->instance.name != NULL) FreeString(ob->instance.name);
        FREE(ob);
        oblast->next = obnext;
     }
@@ -1171,13 +1270,13 @@ static char *OldNodeName(struct nlist *tp, int node)
   struct objlist *firstuniqueglobal;
   struct objlist *firstglobal;
   struct objlist *firstpin;
-  static char StrBuffer[100];
+  static char StrBuffer[MAX_STR_LEN];
 
 #if 0
   /* make second pass, looking for ports */
   ob = tp->cell;
   while (ob != NULL) {
-    if ((node == ob->node) && IsPort(ob->type)) {
+    if ((node == ob->node) && IsPort(ob)) {
       strcpy(StrBuffer, ob->name);
       return(StrBuffer);
     }
@@ -1237,7 +1336,7 @@ static char *OldNodeName(struct nlist *tp, int node)
   for (ob = tp->cell; ob != NULL; ob = ob->next) {
     if (node == ob->node) {
       if (ob->type >= FIRSTPIN) firstpin = ob;
-      else if (IsPort(ob->type)) {
+      else if (IsPort(ob)) {
 	firstport = ob;
 	strcpy(StrBuffer,ob->name);
 	return(StrBuffer);
@@ -1269,7 +1368,6 @@ static char *OldNodeName(struct nlist *tp, int node)
 char *NodeName(struct nlist *tp, int node)
 {
   if (node == -1) return("Disconnected");
-#ifdef CACHE_NODENAMES
   if (tp->nodename_cache != NULL) {
     if (node > tp->nodename_cache_maxnodenum || 
 	tp->nodename_cache[node] == NULL)
@@ -1277,7 +1375,6 @@ char *NodeName(struct nlist *tp, int node)
     else
       return (tp->nodename_cache[node]->name);
   }
-#endif
   return (OldNodeName(tp, node));
 }
 
@@ -1290,34 +1387,30 @@ char *NodeAlias(struct nlist *tp, struct objlist *ob)
 /*    Fprintf(stderr,"Disconnected node in NodeAlias: %s\n",ob->name); */
     return(ob->name);
   }
-#ifdef CACHE_NODENAMES
   if ((ob->node >= 0) && (tp->nodename_cache != NULL) &&
 		(ob->node <= tp->nodename_cache_maxnodenum))
     return (tp->nodename_cache[ob->node]->name);
-#endif
   return (OldNodeName(tp, ob->node));
 }
 
 void FreeNodeNames(struct nlist *tp)
 {
-#ifdef CACHE_NODENAMES
   if (tp == NULL) return;
   if (tp->nodename_cache != NULL)
     FREE(tp->nodename_cache);
   tp->nodename_cache = NULL;
   tp->nodename_cache_maxnodenum = 0;
-#endif
 }
 
 void CacheNodeNames(struct nlist *tp)
 {
-#ifdef CACHE_NODENAMES
   int nodes;
   struct objlist *ob;
 
   if (tp == NULL) return;
   if (tp->nodename_cache != NULL) FreeNodeNames(tp);
   nodes = 0;
+
   for (ob = tp->cell; ob != NULL; ob = ob->next)
     if (ob->node > nodes) nodes = ob->node;
 
@@ -1361,6 +1454,5 @@ void CacheNodeNames(struct nlist *tp)
       break;
     }
   }
-#endif
 }
 
